@@ -6,12 +6,13 @@ module Main (main) where
 import Control.Monad
 import Control.Monad.Catch
 import Control.Monad.IO.Class (MonadIO (..))
+import Data.Bifunctor
 import Data.List (sort)
 import Path
 import Path.IO
-import Test.Hspec
 import System.Environment
 import System.PosixCompat.Files
+import Test.Hspec
 
 #if !MIN_VERSION_base(4,8,0)
 import Control.Applicative ((<$>))
@@ -24,7 +25,9 @@ main = hspec . around withSandbox $ do
     -- NOTE These tests shall fail on Windows as unix-compat does not
     -- implement createSymbolicLink for windows.
     describe "listDir"          listDirSpec
+    describe "listDirRel"       listDirRelSpec
     describe "listDirRecur"     listDirRecurSpec
+    describe "listDirRecurRel"  listDirRecurRelSpec
     describe "listDirRecurWith" listDirRecurWithSpec
     describe "walkDir Finish"   walkDirFinishSpec
     describe "copyDirRecur"     copyDirRecurSpec
@@ -36,6 +39,7 @@ main = hspec . around withSandbox $ do
   describe "getCurrentDir"    getCurrentDirSpec
   describe "setCurrentDir"    setCurrentDirSpec
   describe "withCurrentDir"   withCurrentDirSpec
+  describe "walkDirRel"       walkDirRelSpec
 #ifndef mingw32_HOST_OS
   -- NOTE We can't quite test this on Windows as well, because the
   -- environmental variables HOME and TMPDIR do not exist there.
@@ -52,10 +56,18 @@ listDirSpec :: SpecWith (Path Abs Dir)
 listDirSpec = it "lists directory" $ \dir ->
   getDirStructure listDir dir `shouldReturn` populatedDirTop
 
+listDirRelSpec :: SpecWith (Path Abs Dir)
+listDirRelSpec = it "lists directory" $ \dir ->
+  getDirStructureRel listDirRel dir `shouldReturn` populatedDirTop
+
 listDirRecurSpec :: SpecWith (Path Abs Dir)
 listDirRecurSpec = it "lists directory recursively" $ \dir ->
   getDirStructure listDirRecur dir `shouldReturn` populatedDirStructure
 
+listDirRecurRelSpec :: SpecWith (Path Abs Dir)
+listDirRecurRelSpec = it "lists directory recursively" $ \dir ->
+  getDirStructureRel listDirRecurRel dir `shouldReturn` populatedDirStructure
+
 listDirRecurWithSpec :: SpecWith (Path Abs Dir)
 listDirRecurWithSpec =
   it "lists directory recursively using predicates" $ \dir ->
@@ -75,18 +87,6 @@ listDirRecurWith dirPred filePred =
     f' <- filterM filePred f
     return (d', f')
 
--- Follows symbolic links
-listDirRecurCyclic :: (MonadIO m, MonadThrow m)
-  => Path b Dir                          -- ^ Directory to list
-  -> m ([Path Abs Dir], [Path Abs File]) -- ^ Sub-directories and files
-listDirRecurCyclic = walkDirAccum Nothing (\_ d f -> return (d, f))
-
-listDirRecurCyclicSpec :: SpecWith (Path Abs Dir)
-listDirRecurCyclicSpec =
-  it "lists directory trees having traversal cycles" $ \dir ->
-    getDirStructure listDirRecurCyclic dir
-        `shouldReturn` populatedCyclicDirStructure
-
 -- | walkDir with a Finish handler may have unpredictable output depending on
 -- the order of traversal. The only guarantee is that we will finish only after
 -- we find the directory "c". Though if we test only for the presence of "c" we
@@ -142,6 +142,18 @@ findFileSpec = it "finds a file lazily" $ \dir -> do
   found <- findFile (dir : undefined) relFile
   found `shouldBe` Just (dir </> relFile)
 
+listDirRecurCyclicSpec :: SpecWith (Path Abs Dir)
+listDirRecurCyclicSpec =
+  it "lists directory trees having traversal cycles" $ \dir ->
+    getDirStructure listDirRecurCyclic dir
+        `shouldReturn` populatedCyclicDirStructure
+
+-- Follows symbolic links
+listDirRecurCyclic :: (MonadIO m, MonadThrow m)
+  => Path b Dir                          -- ^ Directory to list
+  -> m ([Path Abs Dir], [Path Abs File]) -- ^ Sub-directories and files
+listDirRecurCyclic = walkDirAccum Nothing (\_ d f -> return (d, f))
+
 getCurrentDirSpec :: SpecWith (Path Abs Dir)
 getCurrentDirSpec = it "returns current dir" $ \dir ->
   getCurrentDir `shouldNotReturn` dir
@@ -160,6 +172,15 @@ withCurrentDirSpec = it "temporarily modifies current dir" $ \dir -> do
     getCurrentDir `shouldReturn` dir
   getCurrentDir `shouldNotReturn` dir
 
+walkDirRelSpec :: SpecWith (Path Abs Dir)
+walkDirRelSpec = it "does not throw exceptions" $ \dir -> do
+  let handler curdir subdirs files = do
+        curdir `shouldBe` $(mkRelDir ".")
+        subdirs `shouldBe` []
+        files `shouldBe` []
+        return WalkFinish
+  walkDirRel handler dir
+
 getHomeDirSpec :: SpecWith (Path Abs Dir)
 getHomeDirSpec =
   it "home dir is influenced by environment variable HOME" $ \dir ->
@@ -257,6 +278,17 @@ getDirStructure f path = do
   rfiles <- sort <$> mapM (makeRelative path) files
   return (rdirs, rfiles)
 
+-- | A version of 'getDirStructure' that accepts scanning function that
+-- returns relative paths.
+
+getDirStructureRel
+  :: (Path Abs Dir -> IO ([Path Rel Dir], [Path Rel File]))
+     -- ^ Which function to use for scanning
+  -> Path Abs Dir
+     -- ^ Path to directory to scan
+  -> IO ([Path Rel Dir], [Path Rel File])
+getDirStructureRel f path = bimap sort sort <$> f path
+
 -- | Structure of directory created by the 'populatedDir' function. Please
 -- keep it sorted.
 
